/*
 * Copyright (C) 2005,2007 Keith Stribley
 * 
 * Distributable under the terms of either the Common Public License or the
 * GNU Lesser General Public License, as specified in the LICENSING.txt file.
 *
 */

package org.sil.graphite;
import java.util.LinkedList;
import java.util.Map;
import java.util.TreeMap;



/**
 * GraphiteFont class to wrap around the C++ graphite font library.
 */
public class GraphiteFont
{
  /**
     * 
     */
    private static final long serialVersionUID = -368561389268636886L;
    private static final String LIBRARY = "graphitejni";
    private static final String LIB_EXT_LINUX = ".so";
    
  // specify some common library locations to try
  private static final String VERSION = "0.0.0";
  private static final String [] ALT_LIBRARY_LINUX = 
  { 
    "lib" + LIBRARY + LIB_EXT_LINUX,
    "/usr/local/lib/lib" + LIBRARY + LIB_EXT_LINUX + "." + VERSION ,
    "/usr/lib/lib" + LIBRARY + LIB_EXT_LINUX + "." + VERSION ,
    "/usr/local/lib/lib" + LIBRARY + LIB_EXT_LINUX, 
    "/usr/lib/lib" + LIBRARY + LIB_EXT_LINUX
  };
  
  private final static int CACHE_SIZE = 20;
  private LinkedList runList = new LinkedList();
  private Map segMap = new TreeMap();
  
  private static boolean libraryLoaded = false;
  /**
   * Static function to load the library - it looks in the standard locations
   */
  synchronized private static void loadLibrary()
  {
    if (libraryLoaded) return;
    try
    {
      try 
      {
        System.loadLibrary(LIBRARY);
        libraryLoaded = true;
      }
      catch (UnsatisfiedLinkError e)
      {
          StringBuilder errorBuilder = new StringBuilder();
          errorBuilder.append(e.getMessage());
          errorBuilder.append(" Failed to load library ");
          String libraryName = System.mapLibraryName(LIBRARY);
          errorBuilder.append(System.mapLibraryName(LIBRARY));
          String [] altLibrary = ALT_LIBRARY_LINUX;
          if (System.getProperty("os.name").toLowerCase().contains("windows"))
          {
              String sep = System.getProperty("path.separator");
              altLibrary = System.getProperty("java.library.path").split(sep);
          }
        for (int i = 0; i<altLibrary.length; i++)
        {
            String testName = altLibrary[i];
            try 
            {
              
              if (!altLibrary[i].contains(libraryName))
              {
                  testName += System.getProperty("file.separator") + libraryName; 
              }
              System.load(testName);
              
              System.out.println("Loaded: " + testName);
              libraryLoaded = true;
              break;
            }
            catch (UnsatisfiedLinkError le)
            {
              errorBuilder.append(le.getMessage());
              errorBuilder.append(" Failed to load library ");
              errorBuilder.append(testName);
              errorBuilder.append(System.getProperty("line.separator"));
            }
        }
        if (!libraryLoaded)
            System.err.println(errorBuilder.toString());
      }
    }
    catch (Exception e)
    {
        e.printStackTrace();
    }
  }
  static {
    if (!libraryLoaded) loadLibrary();    
  }
  
  //private transient long segmentHandle = 0;
  private transient long fontHandle = 0;
  private boolean checkedGraphite = false;
  private boolean isGraphite = false;
  private String fontFileName = null;
  private float size;
  private int dpi; 
  
  /** Creates a new instance of GraphiteFont */
  public GraphiteFont(String fontFileName, float size, int dpi)
  {
    this.fontFileName = fontFileName;
    this.size = size;
    this.dpi = dpi;
    loadLibrary();
    if (!libraryLoaded) System.out.println("No Graphite library found!");
    createGraphiteFont(fontFileName, size, dpi, 400, 0);
  }
  /**
   * Tests if the specified font file has graphite tables
   * @param fontFileName
   * @return
   */
  public boolean isGraphiteFont(String fontFileName)
  {
      GraphiteFont gf = new GraphiteFont(fontFileName, 12, 96);
      return gf.isGraphiteFont();
  }
  /**
   * Checkes whether this font instance has graphite tables.
   */
  public boolean isGraphiteFont()
  {
    if (!checkedGraphite)
    {
      loadLibrary();
      isGraphite = false;
      
      if (fontHandle == 0)
      {
          createGraphiteFont(fontFileName, size, dpi, 400, 0);
      }
        
        if (fontHandle != 0)
        {
          isGraphite = isGraphiteFont(fontHandle);
          checkedGraphite = true;
        }
        //System.out.println("table size " + GraphiteTTF.getTTFTableSize(ttf, ttf.GSUBTag));
      
    }
    return isGraphite;
  }
  
  /**
   * deletes any native handles in the JNI
   */
  public void destroy()
  {
    if (fontHandle != 0)
    {
      destroyFont(fontHandle);
      fontHandle = 0;
      checkedGraphite = false;
    }
  }
  
  /** create a gr::Font object in the JNI 
   * @param path to font file
   * @return handle to the native gr::Font object
   */
  native protected long createGraphiteFont(String fontFile, float pointSize, int dpi);
  /** test the font for the presense of Sile tables */
  native protected boolean isGraphiteFont(long fontHandle);
  /** create a segment from the specified text.
   * @param fontHandle
   * @param text
   * @param isRtl - true if right to left, false otherwise
   * @return handle to the gr::Segment object
   */
  native protected long createSegment(long fontHandle, char [] text, boolean isRtl);
  /**
   * Count the number of glyphs in the segment
   * @param fontHandle
   * @param segmentHandle
   */
  native protected int countGlyphs(long fontHandle, long segmentHandle);
  /**
   * Retrieves the information to create a GlyphVector from a segment.
   * The gid array must be the same size as returned by countGlyphs
   * The xPos and yPos arrays must one bigger than the gid array. The last 
   * index is used to store the advance of the segment.
   * @param fontHandle from createGraphiteFont
   * @param segmentHandle from createSegment
   * @param gid array to be populated with glyph IDs 
   * @param pos array to be populated alternately with glyph x,y positions
   * @param charIndices array to be populated with the char ids corresponding to
   * each glyph. Several glyphs may have the same character index, which will
   * be the first character of each cluster.
   */
  native protected boolean getGlyphArray(long fontHandle, long segmentHandle,
      int [] gid, float [] pos, int [] charIndices);
  
  /**
   * Get the glyph info for a specific glyph in the segment.
   * @param fontHandle
   * @param segmentHandle
   * @param gid
   * @param bools
   * @param ints
   * @param floats
   * @return
   */
  native protected boolean getGlyphInfo(long fontHandle, long segmentHandle, 
          int gid, boolean [] bools, int [] ints, float [] floats);
  
  /**
   * Retreive the glyph info for a specific glyph
   * @param text
   * @param start offset
   * @param end offset
   * @param gid - glyph index of interest
   * @return GlyphInfo object
   */
  public GlyphInfo getGlyphInfo(char [] text, int start, int end, int gid, boolean isRtl)
  {
      long segHandle = getSegHandle(text, start, end, isRtl);
      GlyphInfo gi = new GlyphInfo();
      if (getGlyphInfo(fontHandle, segHandle, gid, gi.bools, gi.ints, gi.floats))
          return gi;
      return null;
  }
  
  /**
   * Deletes the memory allocated to the specified segment. 
   * The segment handle must not be used after this has been called.
   * @param  fontHandle to destroy 
   */
  native protected void destroySegment(long segmentHandle);
  /**
   * Deletes the memory allocated to the specified font. 
   * The font handle must not be used after this has been called.
   * @param  fontHandle to destroy 
   */
  native protected void destroyFont(long fontHandle);
  /**
   * The font's ascent
   */
  public float getAscent() 
  {
    assert(fontHandle != 0);
    return getAscent(fontHandle);
  }
  /**
   * Native method to get the ascent of the given font handle
   * @param fontHandle
   */
  native protected float getAscent(long fontHandle);
  public float getDescent() 
  {
    assert(fontHandle != 0);
    return getDescent(fontHandle);
  }
  /**
   * Native method to get the descent of the given font handle
   * @param fontHandle
   */
  native protected float getDescent(long fontHandle);
  /**
   * Native method to get the height of the given font handle
   */
  public float getHeight()
  {
    assert(fontHandle != 0);
    return getHeight(fontHandle);
  }
  /**
   * Native method to get the ascent of the given font handle
   * @param fontHandle
   */
  native protected float getHeight(long fontHandle);
  /**
   * Construct an native font object
   * @param fontFile
   * @param pointSize
   * @param dpi - dots per inch
   * @param weight - bold?
   * @param slant - italic?
   * @return pointer to native font object
   */
  public void createGraphiteFont(String fontFile, float pointSize, int dpi, int weight, int slant)
  {
      fontHandle = createGraphiteFont(fontFile, pointSize, dpi);
      // TODO error handling, bold, italic
  }
  /** create a segment 
   * @param fontHandle
   * @param text
   * @return handle to the gr::Segment object
   */
  public void createSegment(char [] text, int start, int end, boolean isRtl)
  {
      if (runList.size() == CACHE_SIZE)
      {
          String oldestKey = runList.removeFirst().toString();
          if (oldestKey != null)
          {
              long oldestHandle = ((Long)segMap.get(oldestKey)).longValue();
              destroySegment(oldestHandle);
              segMap.remove(oldestKey);
          }
      }
      String run = new String(text, start, end - start);
      if (!segMap.containsKey(run))
      {
          long segHandle = createSegment(fontHandle, run.toCharArray(), isRtl);
          segMap.put(run, new Long(segHandle));
          runList.addLast(run);
      }
//    TODO error handling
  }
  /**
   * Creates a segment from the specified text.
   * @param fontHandle
   * @param segmentHandle
   */
  public int countGlyphs(char [] text, int start, int end, boolean isRtl)
  {
      if (end == start) return 0;
      String run = new String(text, start, end - start);
      long segHandle = 0;
      if (segMap.containsKey(run))
      {
          segHandle = ((Long)segMap.get(run)).longValue();
      }
      else
      {
          createSegment(text, start, end, isRtl);
          if (segMap.containsKey(run))
          {
              segHandle = ((Long)segMap.get(run)).longValue();
          }
          else return 0;
      }
      return countGlyphs(fontHandle, segHandle);
  }
  
  
  public float getWidth(char [] text, int start, int end, boolean isRtl)
  {
      long segHandle = 0;
      String run = new String(text, start, end - start);
      if (segMap.containsKey(run))
      {
          segHandle = ((Long)segMap.get(run)).longValue();
      }
      else
      {
          createSegment(text, start, end, isRtl);
          segHandle = ((Long)segMap.get(run)).longValue();
      }
      int glyphCount = countGlyphs(fontHandle, segHandle);
      int [] gid = new int [glyphCount];
      float [] pos = new float [2 * glyphCount + 2];
      int [] charIndices = new int [end - start];
      if (getGlyphArray(fontHandle, segHandle, gid, pos, charIndices))
      {
          return pos[2 * glyphCount];
      }
      return 0;
  }
  
  /**
   * Retrieves the information to create a GlyphVector from a segment.
   * The gid array must be the same size as returned by countGlyphs
   * The xPos and yPos arrays must one bigger than the gid array. The last 
   * index is used to store the advance of the segment.
   * @param fontHandle from createGraphiteFont
   * @param segmentHandle from createSegment
   * @param gid array to be populated with glyph IDs 
   * @param pos array to be populated alternately with glyph x,y positions
   * @param charIndices array to be populated with the char ids corresponding to
   * each glyph
   */
  public boolean getGlyphArray(char [] text, int start, int end,
      int [] gid, float [] pos, int [] charIndices, boolean isRtl)
  {
      long segHandle = getSegHandle(text, start, end, isRtl);
      // TODO handle context properly
      return getGlyphArray(fontHandle, segHandle, gid, pos, charIndices);
  }
  
  protected long getSegHandle(char [] text, int start, int end, boolean isRtl)
  {
      String run = new String(text, start, end - start);
      long segHandle = 0;
      if (segMap.containsKey(run))
      {
          segHandle = ((Long)segMap.get(run)).longValue();
      }
      else
      {
          createSegment(text, start, end, isRtl);
          segHandle = ((Long)segMap.get(run)).longValue();
      }
      return segHandle;
  }
  
  /**
   * Deletes the memory allocated to the specified font. 
   * The font handle must not be used after this has been called.
   * @param  fontHandle to destroy 
   */
  
  public void dispose()
  {
    if (fontHandle != 0) 
    {      
      destroyFont(fontHandle);
      fontHandle = 0;
    }
    //System.out.println("GraphiteFont " + fontFileName + " disposed");
  }
  
  // offsets for glyph info
      static final int GI_BOOL_IS_SPACE = 0;
      static final int GI_BOOL_INSERT_BEFORE = 1;
      static final int GI_BOOL_IS_ATTACHED = 2;
      static final int GI_BOOL_ERRONEOUS = 3;
      static final int GI_BOOL_MAX = 4;
      
      static final int GI_INT_GLYPH_ID = 0;
      static final int GI_INT_PSEUDO_GLYPH_ID = 1;
      static final int GI_INT_LOGICAL_INDEX = 2;
      static final int GI_INT_FIRST_CHAR = 3;
      static final int GI_INT_LAST_CHAR = 4;
      static final int GI_INT_DIRECTIONALITY = 5;
      static final int GI_INT_DIRECTION_LEVEL = 6;
      static final int GI_INT_BREAK_WEIGHT = 7;
      static final int GI_INT_MAX = 8;

      static final int GI_FLOAT_ORIGIN = 0;
      static final int GI_FLOAT_ADVANCE_WIDTH = 1;
      static final int GI_FLOAT_ADVANCE_HEIGHT = 2;
      static final int GI_FLOAT_Y_OFFSET = 3;
      static final int GI_FLOAT_BB_LEFT = 4;
      static final int GI_FLOAT_BB_RIGHT = 5;
      static final int GI_FLOAT_BB_TOP = 6;
      static final int GI_FLOAT_BB_BOTTOM = 7;
      static final int GI_FLOAT_MAX = 8;
}
