/*
 * Copyright (C) 2007 Keith Stribley
 * 
 * Distributable under the terms of either the Common Public License or the
 * GNU Lesser General Public License, as specified in the LICENSING.txt file.
 *
 */
package org.sil.graphite.swing;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.Timer;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.plaf.TextUI;
import javax.swing.text.BadLocationException;
import javax.swing.text.Caret;
import javax.swing.text.DefaultCaret;
import javax.swing.text.Document;
import javax.swing.text.Highlighter;
import javax.swing.text.JTextComponent;
import javax.swing.text.Position;
import javax.swing.text.View;
import javax.swing.text.Position.Bias;

/**
 * @author keith
 *
 */
public class GrCaret extends DefaultCaret
{
    private static final long serialVersionUID = -7340928074994666137L;
    final static int BLINK_RATE = 250;
    GraphiteGlyphPainter ggp = null;
    // unfortunately most of the equivalent variables in the base class are
    // private and the setters do unwanted stuff
    private int grDot = 0;
    private int grMark = 0;
    private double xOffset = 0.0;
    private boolean grVisible = false;
    private Position.Bias grDotBias = Position.Bias.Forward;
    private Position.Bias grMarkBias = Position.Bias.Forward;
    private Rectangle caretRect = null;
    private BlinkHandler handler = null;
    private Timer flasher = null;
    private JTextComponent textComponent = null;
    private Object selectionTag = null;
    private Rectangle repaintRect = null;
    /* (non-Javadoc)
     * @see javax.swing.text.DefaultCaret#getDotBias()
     */
    @Override
    public Bias getDotBias()
    {
        return grDotBias;
    }
    /* (non-Javadoc)
     * @see javax.swing.text.DefaultCaret#getMarkBias()
     */
    @Override
    public Bias getMarkBias()
    {
        return grMarkBias;
    }
    
    public GrCaret(GraphiteGlyphPainter p)
    {
        this.ggp = p;
        this.handler = new BlinkHandler(this);
    }
    public void setPainter(GraphiteGlyphPainter p)
    {
        this.ggp = p;
    }

    /* (non-Javadoc)
     * @see javax.swing.text.DefaultCaret#setDot(int, javax.swing.text.Position.Bias)
     */
    @Override
    public void setDot(int dot, Bias dotBias)
    {
        if (textComponent == null) 
            return;
        if (dot != grDot)
        {
            int direction = (dot > grDot) ? View.EAST : View.WEST;
            grDot = ggp.getNextValidCaretPos(ggp.getView(), dot, direction);
            grDotBias = dotBias;
            grMark = grDot;
            grMarkBias = dotBias;
            textComponent.getHighlighter().removeAllHighlights();
            computeCaret();
        }
    }
    
    protected void computeCaret()
    {
        
        Rectangle oldRect = null;
        if (caretRect != null)
            oldRect = (Rectangle)caretRect.clone();
        if (!getCaretRectangle().equals(oldRect))
        {
            repaintRect = new Rectangle(0,0, 
                textComponent.getWidth(), textComponent.getHeight());
            if (textComponent.isFocusOwner())
            {
                this.setBlinkRate(250);
                grVisible = true;
            }
        }
        else if (grDot != grMark)
        {
            repaintRect = new Rectangle(0,0, 
                    textComponent.getWidth(), textComponent.getHeight());
        }
        else
        {
            repaintRect = caretRect;
        }
    }

    protected Rectangle getCaretRectangle()
    {
        if (textComponent == null)
            return null;
        Rectangle rect = new Rectangle(0, 0, textComponent.getWidth(), 
                textComponent.getHeight());
        try
        {
            Shape dotShape = ggp.modelToView(ggp.getView(), grDot, grDotBias, rect);
            if (dotShape instanceof Rectangle)
            {
                caretRect = (Rectangle)dotShape;
                setMagicCaretPosition(new Point(caretRect.x, caretRect.y));
                FontMetrics fm = textComponent.getFontMetrics(textComponent.getFont());
                caretRect.x += textComponent.getInsets().left;
                caretRect.y += textComponent.getInsets().top;
                int yOffset = (textComponent.getHeight() - fm.getHeight() 
                        - textComponent.getInsets().top 
                        - textComponent.getInsets().bottom) / 2;
                caretRect.y += yOffset;
                caretRect.width = 1;
                caretRect.height = fm.getHeight();
                
                if (grMark == grDot && selectionTag != null)
                {
                    textComponent.getHighlighter().removeHighlight(selectionTag);
                    selectionTag  = null;
                }
            }
                
        }
        catch (BadLocationException e)
        {
            e.printStackTrace();
            grDot = 0;
        }
        return caretRect;
    }

    /* (non-Javadoc)
     * @see javax.swing.text.DefaultCaret#moveDot(int, javax.swing.text.Position.Bias)
     */
    @Override
    public void moveDot(int dot, Bias dotBias)
    {
        if (textComponent == null) 
            return;
        if (dot != grDot)
        {
            int direction = (dot > grDot) ? View.EAST : View.WEST;
            grDot = ggp.getNextValidCaretPos(ggp.getView(), dot, direction);
            grDotBias = dotBias;
            computeCaret();
            if (grDot != grMark)
            {
                int p0 = Math.min(grDot, grMark);
                int p1 = Math.max(grDot, grMark);
                try
                {
                    textComponent.getHighlighter().addHighlight(p0, p1, 
                            getSelectionPainter());
                } 
                catch (BadLocationException e)
                {
                    e.printStackTrace();
                }
            }
            else
            {
                textComponent.getHighlighter().removeAllHighlights();
            }
        }
    }
    
    
    /* (non-Javadoc)
     * @see javax.swing.text.DefaultCaret#getDot()
     */
    @Override
    public int getDot()
    {
        return grDot;
    }
    /* (non-Javadoc)
     * @see javax.swing.text.DefaultCaret#getMark()
     */
    @Override
    public int getMark()
    {
        return grMark;
    }
    /* (non-Javadoc)
     * @see javax.swing.text.DefaultCaret#paint(java.awt.Graphics)
     */
    @Override
    public void paint(Graphics g)
    {
        if (g instanceof Graphics2D)
        {
            Graphics2D g2d = (Graphics2D)g;
            if (caretRect != null)
            {
                this.x = caretRect.x - (int)xOffset;
                this.y = caretRect.y;
                this.width = caretRect.width;
                this.height = caretRect.height;
                
                g2d.setColor(Color.BLACK);
                g2d.fill(this);
                
                
                //System.out.println("GrCaret paint" + this.x);
            }
        }
    }
    /* (non-Javadoc)
     * @see javax.swing.text.DefaultCaret#deinstall(javax.swing.text.JTextComponent)
     */
    @Override
    public void deinstall(JTextComponent c)
    {
        c.removePropertyChangeListener(handler);
        c.removeFocusListener(handler);
        c.removeMouseListener(handler);
        Document doc = c.getDocument();
        if (doc != null) {
            doc.removeDocumentListener(handler);
        }
        textComponent = null;
    }
    /* (non-Javadoc)
     * @see javax.swing.text.DefaultCaret#install(javax.swing.text.JTextComponent)
     */
    @Override
    public void install(JTextComponent c)
    {
        Document doc = c.getDocument();
        textComponent = c;
        if (doc != null) {
            doc.addDocumentListener(handler);
        }
        c.addPropertyChangeListener(handler);
        c.addFocusListener(handler);
        c.addMouseListener(handler);
    }

    /**
     * Sets the caret blink rate.
     *
     * @param rate the rate in milliseconds, 0 to stop blinking
     * @see Caret#setBlinkRate
     */
    public void setBlinkRate(int rate) {
        if (rate != 0) 
        {
            if (flasher == null) 
            {
                flasher = new Timer(rate, handler);
                flasher.start();
            }
            else
            {
                flasher.setDelay(rate);
                if (!flasher.isRunning())
                    flasher.start();
            }
        } 
        else 
        {
            if (flasher != null) 
            {
                flasher.stop();
                flasher.removeActionListener(handler);
                flasher = null;
            }
        }
    }

    
    class BlinkHandler implements PropertyChangeListener, DocumentListener, 
        ActionListener, FocusListener, MouseListener
    {
        GrCaret caret = null;
        public BlinkHandler(GrCaret caret)
        {
            this.caret = caret;
        }
        /* (non-Javadoc)
         * @see java.beans.PropertyChangeListener#propertyChange(java.beans.PropertyChangeEvent)
         */
        public void propertyChange(PropertyChangeEvent evt)
        {
            
        }

        /* (non-Javadoc)
         * @see javax.swing.event.DocumentListener#changedUpdate(javax.swing.event.DocumentEvent)
         */
        public void changedUpdate(DocumentEvent e)
        {
            
        }

        /* (non-Javadoc)
         * @see javax.swing.event.DocumentListener#insertUpdate(javax.swing.event.DocumentEvent)
         */
        public void insertUpdate(DocumentEvent e)
        {
            
        }

        /* (non-Javadoc)
         * @see javax.swing.event.DocumentListener#removeUpdate(javax.swing.event.DocumentEvent)
         */
        public void removeUpdate(DocumentEvent e)
        {
            
        }

        /* (non-Javadoc)
         * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
         */
        public void actionPerformed(ActionEvent e)
        {
            if (e.getSource() == flasher)
            {
                textComponent.repaint(repaintRect);
                // toggle visibility
                if (grVisible)
                {
                    grVisible = false;
                }
                else if (textComponent.isFocusOwner())
                {
                    grVisible = true;
                }
            }
        }
        /* (non-Javadoc)
         * @see java.awt.event.FocusListener#focusGained(java.awt.event.FocusEvent)
         */
        public void focusGained(FocusEvent e)
        {
            setBlinkRate(BLINK_RATE);
            
        }
        /* (non-Javadoc)
         * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
         */
        public void focusLost(FocusEvent e)
        {
            setBlinkRate(0);
            grVisible = false;
            textComponent.repaint(caret);
        }
        /* (non-Javadoc)
         * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
         */
        public void mouseClicked(MouseEvent e)
        {
            
        }
        /* (non-Javadoc)
         * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
         */
        public void mouseEntered(MouseEvent e)
        {
            // TODO Auto-generated method stub
            
        }
        /* (non-Javadoc)
         * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
         */
        public void mouseExited(MouseEvent e)
        {
            // TODO Auto-generated method stub
            
        }
        /* (non-Javadoc)
         * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
         */
        public void mousePressed(MouseEvent e)
        {
            if (!textComponent.isFocusOwner())
                textComponent.grabFocus();
            Point click = e.getPoint();
            checkView();
            Bias [] biasReturn = { Bias.Forward };
            int newPos = ggp.viewToModel(ggp.getView(), click.x, click.y, 
                                e.getComponent().getBounds(), biasReturn);
            grMark = newPos;
        }

        /* (non-Javadoc)
         * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
         */
        public void mouseReleased(MouseEvent e)
        {
            Point click = e.getPoint();
            
            Bias [] biasReturn = { Bias.Forward };
            checkView();
            int newPos = ggp.viewToModel(ggp.getView(), click.x, click.y, 
                                e.getComponent().getBounds(), biasReturn);
            grDot = newPos;
            computeCaret();
            Highlighter h = textComponent.getHighlighter();
            if (grDot != grMark)
            {
                int p0 = Math.min(grDot, grMark);
                int p1 = Math.max(grDot, grMark);
                Highlighter.HighlightPainter p = getSelectionPainter();
                
                try
                {
                    selectionTag = h.addHighlight(p0, p1, p);
                } 
                catch (BadLocationException e1)
                {
                    e1.printStackTrace();
                }
            }
            else
            {
                if (selectionTag != null)
                {
                    h.removeHighlight(selectionTag);
                    selectionTag  = null;
                }
            }
        }
        
    }

    private void checkView()
    {
        if (textComponent.getUI() instanceof TextUI)
        {
            TextUI textUI = (TextUI)textComponent.getUI();
            if (textUI.getRootView(textComponent).getViewCount() > 0)
            {
                View v = textUI.getRootView(textComponent).getView(0);
                if (v != ggp.getView())
                {
                    textComponent.repaint();
                }
            }
        }
    }

    /* (non-Javadoc)
     * @see javax.swing.text.DefaultCaret#isVisible()
     */
    @Override
    public boolean isVisible()
    {
        return grVisible;
    }
    /* (non-Javadoc)
     * @see javax.swing.text.DefaultCaret#setVisible(boolean)
     */
    @Override
    public void setVisible(boolean visible)
    {
        grVisible = visible;
    }
    
    public void setXOffset(double xOffset)
    {
        this.xOffset = xOffset; 
    }
}
