/*
 * The GlyphPainter interface is Copyright 2006 Sun Microsystems, Inc. 
 * All rights reserved.
 * 
 * This implementation for Graphite is Copyright (C) 2007 Keith Stribley 
 *  
 * Distributable under the terms of either the Common Public License or the
 * GNU Lesser General Public License, as specified in the LICENSING.txt file.
 * 
 */

package org.sil.graphite.swing;

import java.text.DecimalFormat;
import java.awt.Font;
import java.awt.Shape;
import java.awt.Point;
import java.awt.geom.Point2D;
import java.awt.FontMetrics;
import java.awt.Rectangle;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Container;
import java.awt.RenderingHints;
import java.awt.font.FontRenderContext;
import java.awt.font.GlyphVector;

import javax.swing.text.JTextComponent;
import javax.swing.text.Caret;
import javax.swing.text.View;
import javax.swing.text.Position;
import javax.swing.text.BadLocationException;
import javax.swing.text.GlyphView;
import javax.swing.text.GlyphView.GlyphPainter;
import javax.swing.text.TabExpander;

import org.sil.graphite.awt.GraphiteAwtFont;

/**
 * A class to perform rendering of the glyphs for Graphite fonts.
 * This is designed as a drop in replacement for a GlyphView GlyphPainter
 */
public class GraphiteGlyphPainter extends javax.swing.text.GlyphView.GlyphPainter
{
    private GraphiteAwtFont gFont = null;
    private View view = null;
    private Graphics2D g2d = null;
    private GlyphVector cachedGlyphVector = null;
    private String cachedText = null;
    private int showError = 10; // show upto this number of errors
    private boolean debug = false;
    private int p0; 
    private int p1;
    private String lastMsg = ""; // for debugging
    private GlyphPainter origPainter = null;
    final static boolean USE_ORIG = false;
    /** Creates a new instance of GraphiteGlyphPainter */
    public GraphiteGlyphPainter(GraphiteAwtFont font, View v, Graphics2D g2d, int p0, int p1)
    {
        this.gFont = font;
        this.view = v;
        this.p0 = p0;
        this.p1 = p1;
        this.g2d = g2d;
    }
      
    public void setOriginalPainter(GlyphPainter p)
    {
        this.origPainter = p;
    }

	/**
	 * Determine the span the glyphs given a start location
	 * (for tab expansion).
     * javax.swing.text.GlyphView.GlyphPainter#getSpan
	 */
	public float getSpan(GlyphView v, int p0, int p1, TabExpander e, float x)
    {
        if (USE_ORIG && origPainter != null)
            return origPainter.getSpan(v, p0, p1, e, x);
        if (gFont == null) return 0.0f;
        String text = v.getText(p0, p1).toString();
        GlyphVector gv = getGlyphVector(text);
        return (float)gv.getGlyphPosition(gv.getNumGlyphs()).getX();
    }

	public float getHeight(GlyphView v)
    {
      if (USE_ORIG && origPainter != null) return origPainter.getHeight(v);
      if (v == null || gFont == null)
          return 12f;
      return v.getContainer().getFontMetrics(gFont).getHeight();
    }

	public float getAscent(GlyphView v)
    {
        if (USE_ORIG && origPainter != null) return origPainter.getAscent(v);
        if (v == null || gFont == null) return 0f;
        return v.getContainer().getFontMetrics(gFont).getAscent();
    }

	public float getDescent(GlyphView v)
    {
        if (USE_ORIG && origPainter != null) return origPainter.getDescent(v);
        if (v == null || gFont == null) return 0f;
        return v.getContainer().getFontMetrics(gFont).getDescent();
    }

	/**
	 * Paint the glyphs representing the given range.
     * javax.swing.text.GlyphView.GlyphPainter#paint
	 */
    public void paint(View v, Graphics g, Shape a, int p0, int p1)
    {
        try
        {
            if (g == null || gFont == null) return;
            
            g2d = (Graphics2D)g;
            //System.out.println("Graphics " + g);
            String text = null;
            g2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, 
                  RenderingHints.VALUE_TEXT_ANTIALIAS_LCD_HRGB);
            //g2d.setColor(v.getContainer().getBackground());
            //g2d.fill(a);
            // update offsets
            this.p0 = p0;
            this.p1 = p1;
            if (v instanceof GlyphView)
            {
                GlyphView glyphView = (GlyphView)v;
                g2d.setColor(glyphView.getForeground());
                text = glyphView.getText(p0, p1).toString();
            }
            else
            {
                g2d.setColor(v.getContainer().getForeground());
                text = v.getDocument().getText(p0, p1).toString();
            }
            GlyphVector gv = getGlyphVector(text);
            if (gv != null)
            {
                Rectangle alloc = getBounds(a);
                FontMetrics fm = v.getContainer().getFontMetrics(gFont);
                g2d.drawGlyphVector(gv, alloc.x , alloc.y + 
                                    fm.getAscent() + fm.getLeading());
                if (v.getStartOffset() != p0 || v.getEndOffset() != p1)
                {
                  if (showError-- > 0)
                  {
                    assertPrint("paint does not match view!");
                  }
                }
            }
            else
            {
                if (showError-- > 0)
                {
                  assertPrint("Null Graphite glyph vector!");
                }
            }
        }
        catch (BadLocationException e)
        {
            e.printStackTrace();
        }
    }

    /**
     * @see javax.swing.text.GlyphView.GlyphPainter#modelToView
     */
    public Shape modelToView(GlyphView v, 
                      int pos, Position.Bias bias,
                      Shape a) throws BadLocationException
    {
        return this.modelToView((View)v, pos, bias, a);
    }
    
	/**
     * A generalisation that allows other views to be used as well as GlyphView
	 * @see javax.swing.text.GlyphView.GlyphPainter#modelToView
	 */
	public Shape modelToView(View v, 
					  int pos, Position.Bias bias,
					  Shape a) throws BadLocationException
    {
        Shape origShape = null;
        this.p0 = v.getStartOffset();
        this.p1 = v.getEndOffset();
        if (origPainter != null && v instanceof GlyphView)
        {
            origShape = origPainter.modelToView((GlyphView)v, pos, bias, a);
            if (USE_ORIG)
            {
                return origShape;
            }
        }
        GlyphVector gv = getGlyphVector(view);
        Rectangle clusterBounds = null;
        Rectangle viewBounds = getBounds(a);
        //int charLength = text.length();
        if (pos > v.getEndOffset()) 
        {
            System.out.println("BadLocation" + pos);
            throw new BadLocationException("modelToView - outside view", pos);
        }
        else if (gv == null || pos == v.getEndOffset() || gv.getNumGlyphs() == 0)
        {
            return new Rectangle(new Point((int)viewBounds.getMaxX(),(int)viewBounds.getY()));
        }
        int localPos = pos - v.getStartOffset();
        int glyphCount = gv.getNumGlyphs();
        int charIndex = gv.getGlyphCharIndex(0);
        clusterBounds = getBounds(gv.getGlyphLogicalBounds(0));
        double rightEdge = 0;
        int i = 0;
        while ((charIndex < localPos) && (i + 1 < glyphCount))
        {
            rightEdge = Math.max(rightEdge, clusterBounds.getX() + clusterBounds.getWidth());
            if (++i < glyphCount)
            {
                charIndex = gv.getGlyphCharIndex(i);
                clusterBounds = getBounds(gv.getGlyphLogicalBounds(i));
            }
            else break;
        }
        clusterBounds = new Rectangle((int)(rightEdge + viewBounds.getX()), 
              (int)viewBounds.getY(), 0, (int)viewBounds.getHeight());
        if (debug)
        {
            // debugging - avoid lots of repeats of same msg
            String msg = "modelToView " + pos + " x " + clusterBounds.getMinX()
                + " " + clusterBounds.getMaxX() + 
                "(" + v.getStartOffset() +","+ v.getEndOffset() + ")" + bias;
            if (!msg.equals(lastMsg))
            {
                assertPrint(msg);
                lastMsg = msg;
            }
        }
        return clusterBounds;
    }

    
    /**
     *  
     *  @see javax.swing.text.GlyphView.GlyphPainter#viewToModel
     */
    public int viewToModel(GlyphView v, 
            float x, float y, Shape a, 
            Position.Bias[] biasReturn)
    {
        return this.viewToModel((View)v, x, y, a, biasReturn);
    }
	/**
     * A generalisation that allows other views to be used as well as GlyphView
	 * @see View#viewToModel
	 */
    public int viewToModel(View v, 
                float x, float y, Shape a, 
                Position.Bias[] biasReturn)
    {
      //String text = v.getText(v.getStartOffset(), v.getEndOffset()).toString();
        this.p0 = v.getStartOffset();
        this.p1 = v.getEndOffset();
      GlyphVector gv = getGlyphVector(view);
      if (gv == null)
      {
          assertPrint("null gv in viewToModel");             
          return v.getStartOffset();
      }
      int prevChar = 0;
      int charLength = v.getEndOffset() - v.getStartOffset();
      int glyphCount = gv.getNumGlyphs();
      int lastFit = 0;
      Rectangle viewBounds = getBounds(a);
      float targetPos = x - ((float)viewBounds.getX());
      for (int i = 0; i < glyphCount; i++)
      {
        prevChar = gv.getGlyphCharIndex(i);
        Rectangle glyphBounds = getBounds(gv.getGlyphLogicalBounds(i));
        assertPrint("char " + prevChar + "glyphbounds " + glyphBounds.getX() + " width "  +
            glyphBounds.getWidth());
        while (i + 1 < glyphCount && prevChar < charLength && 
               gv.getGlyphCharIndex(i + 1) <= prevChar)
        {
          i++;
          glyphBounds = glyphBounds.union(getBounds(gv.getGlyphLogicalBounds(i)));   
          prevChar = gv.getGlyphCharIndex(i);
        }
        if (glyphBounds.getMinX() > targetPos)
        {
          if (i + 1 == glyphCount)
            lastFit = charLength - 1;
          else 
            lastFit = prevChar;
          biasReturn[0] = Position.Bias.Forward;
          break;
        }
        else if (glyphBounds.getMinX() <= targetPos)
        {
          lastFit = prevChar;
          if (targetPos - glyphBounds.getMinX() < 
              glyphBounds.getMaxX() - targetPos)
          {
            biasReturn[0] = Position.Bias.Backward;
            break;
          }
          else if (i + 1 == glyphCount)
          {
            lastFit = charLength - 1;
          }
        }
        else
        {
          // exceeded requested width so break and use lastFit
          break;
        }
      }
      
      int pos = v.getStartOffset() + lastFit;
      assertPrint("viewToModel x" + x + " pox=" + pos + " " + biasReturn.toString());
      return pos;
    }

	/**
	 * @see javax.swing.text.GlyphView.GlyphPainter#breakView
	 */
    public int getBoundedPosition(GlyphView v, int p0, float x, float len)
    {
        if (USE_ORIG && origPainter != null)
        {
            return origPainter.getBoundedPosition(v, p0, x, len);
        }
        if (gFont == null) return p0;
      String text = v.getText(v.getStartOffset(), v.getEndOffset()).toString();
      if (p0 > v.getStartOffset())
      {
        if (showError-- > 0) System.out.println("sub view positioning not yet implemented");
      }
      GlyphVector gv = getGlyphVector(text);
      if (gv == null) 
      {
          assertPrint("null gv");
          return p0;
      }
      int prevChar = 0;
      int charLength = text.length();
      int glyphCount = gv.getNumGlyphs();
      int lastFit = 0;
      int bestPos = 0;
      for (int i = 0; i < glyphCount; i++)
      {
        prevChar = gv.getGlyphCharIndex(i);
        Rectangle glyphBounds = getBounds(gv.getGlyphLogicalBounds(i));
        
        while (i + 1 < glyphCount && prevChar + 1 < charLength && 
               gv.getGlyphCharIndex(i + 1) <= prevChar)
        {
          i++;
          glyphBounds = glyphBounds.union(getBounds(gv.getGlyphLogicalBounds(i)));       
          prevChar = gv.getGlyphCharIndex(i);
        }
        if (glyphBounds.getMaxX() < len)
        {
          if (i + 1 == glyphCount)
          {
            lastFit = charLength - 1;
            bestPos = lastFit;
          }
          else 
            lastFit = prevChar;
        }
        else
        {
          bestPos = lastFit;
          // exceeded requested width so break and use lastFit
          break;
        }
      }
      int pos = v.getStartOffset() + bestPos;
      return pos;
    }

	/**
     * @see javax.swing.text.GlyphView.GlyphPainter#getPainter
	 */
    public GlyphPainter getPainter(GlyphView v, int p0, int p1) 
    {
      return new GraphiteGlyphPainter(gFont, v, g2d, p0, p1);
	}

	/**
     * @see javax.swing.text.GlyphView.GlyphPainter#getNextVisualPositionFrom
	 */
     
    public int getNextVisualPositionFrom(GlyphView v, int pos, Position.Bias b, Shape a, 
					     int direction,
					     Position.Bias[] biasRet) 
	    throws BadLocationException {
        if (USE_ORIG && origPainter != null)
        {
            return origPainter.getNextVisualPositionFrom(v, pos, b, a, direction, biasRet);
        }
	    int startOffset = v.getStartOffset();
	    int endOffset = v.getEndOffset();
	    
	    switch (direction) {
	    case View.NORTH:
	    case View.SOUTH:
                if (pos != -1) {
                    // Presumably pos is between startOffset and endOffset,
                    // since GlyphView is only one line, we won't contain
                    // the position to the north/south, therefore return -1.
                    return -1;
                }
                Container container = v.getContainer();

                if (container instanceof JTextComponent) {
                    Caret c = ((JTextComponent)container).getCaret();
                    Point magicPoint;
                    magicPoint = (c != null) ? c.getMagicCaretPosition() :null;

                    if (magicPoint == null) {
                        biasRet[0] = Position.Bias.Forward;
                        return startOffset;
                    }
                    int value = v.viewToModel(magicPoint.x, 0f, a, biasRet);
                    return value;
                }
                break;
	    case View.EAST:
    		if(startOffset == v.getDocument().getLength()) {
    		    if(pos == -1) {
    		        biasRet[0] = Position.Bias.Forward;
    		        return startOffset;
    		    }
    		    // End case for bidi text where newline is at beginning
    		    // of line.
    		    return -1;
    		}
    		if(pos == -1) {
    		    biasRet[0] = Position.Bias.Forward;
    		    return startOffset;
    		}
    		if(pos == endOffset) {
    		    return -1;
    		}
    		if(++pos == endOffset) {
                // Assumed not used in bidi text, GlyphPainter2 will
                // override as necessary, therefore return -1.
                return -1;
    		}
    		else {
    		    biasRet[0] = Position.Bias.Forward;
                pos = getNextValidCaretPos(v, pos, direction);
                assertPrint("getNextVisualPositionFrom " + pos + 
                    " " + (biasRet[0] == Position.Bias.Forward)); 
    		}
    		return pos;
	    case View.WEST:
    		if(startOffset == v.getDocument().getLength()) {
    		    if(pos == -1) {
    			biasRet[0] = Position.Bias.Forward;
    			return startOffset;
    		    }
    		    // End case for bidi text where newline is at beginning
    		    // of line.
    		    return -1;
    		}
    		if(pos == -1) {
                        // Assumed not used in bidi text, GlyphPainter2 will
                        // override as necessary, therefore return -1.
    		    biasRet[0] = Position.Bias.Forward;
    		    return endOffset - 1;
    		}
    		if(pos == startOffset) {
    		    return -1;
    		}
    		biasRet[0] = Position.Bias.Forward;
            --pos;
            pos = getNextValidCaretPos(v, pos, direction);
            assertPrint("getNextVisualPositionFrom " + pos + 
                    " " + biasRet[0]); 
    		return (pos);
	    default:
	        throw new IllegalArgumentException("Bad direction: " + direction);
	    }
	    return pos;
	}
    
    protected int getNextValidCaretPos(GlyphView v, int pos, int direction)
    {
        return this.getNextValidCaretPos((View)v, pos, direction);
    }
    
    protected int getNextValidCaretPos(View v, int pos, int direction)
    {
        this.p0 = v.getStartOffset();
        this.p1 = v.getEndOffset();
      String text = null;
      if (v instanceof GlyphView)
      {
          text = ((GlyphView)v).getText(v.getStartOffset(), v.getEndOffset()).toString();
      }
      else
      {
          try
          {
              text = v.getDocument().getText(v.getStartOffset(), v.getEndOffset() - v.getStartOffset());
          }
          catch (BadLocationException e)
          {
              e.printStackTrace();
              text = "";
          }
      }
      int localPos = pos - v.getStartOffset();
      GlyphVector gv = getGlyphVector(view);
      int oldPos = pos;
      int validPos = pos;
      if (gv == null) return pos;
      // special case for pos being a space
      if (text.charAt(localPos) == ' ')
        return pos;
      int glyphCount = gv.getNumGlyphs();
      int lastValidPos = -1;
      for (int i = 0; i < glyphCount; i++)
      {
        validPos = gv.getGlyphCharIndex(i);
        while (i + 1 < glyphCount && gv.getGlyphCharIndex(i + 1) < validPos)
        {
          validPos = gv.getGlyphCharIndex(++i);
        }
        if (validPos == localPos)
        {
          break;
        }
        else if (validPos > localPos)
        {
          if (direction == View.EAST)
          {
            // if going forwards then this is the first valid position after the 
            //last requested
            break;
          }
          else
          {
            // going backwards we have overshot, so go back to previous valid
            // position i.e. start of previous cluster
            assert(direction == View.WEST);
            validPos = lastValidPos;
            break;
          }
        }
        lastValidPos = validPos;
      }
      // return to document indices
      validPos += v.getStartOffset();
      //if (validPos == 0 || validPos >= v.getEndOffset()) validPos = -1;
      assertPrint("getNextValidCharPos " + oldPos + "->" + validPos);
      return validPos;
    }
    
    /**
     * Helper to get a glyph vector from the current view
     * @param v - view associated with this painter
     * @return graphite glyph vector
     */
    protected GlyphVector getGlyphVector(View v)
    {
        try
        {
            String text = null; 
            if (v instanceof GlyphView)
            {
                GlyphView glyphView = (GlyphView)v;
                g2d.setColor(glyphView.getForeground());
                text = glyphView.getText(p0, p1).toString();
            }
            else
            {
                if (v.getContainer() == null || v.getDocument() == null) 
                    return null; 
                g2d.setColor(v.getContainer().getForeground());
                text = v.getDocument().getText(p0, p1).toString();
            }
            return getGlyphVector(text);
        }
        catch (BadLocationException e)
        {
            e.printStackTrace();
        }
        return null;
    }
    
    /**
     * Helper to get a glyph vector from a string
     * @param text to render
     * @return graphite glyph vector
     */
    protected GlyphVector getGlyphVector(String text)
    {
        GlyphVector gv = null;
        
        if (text.equals(cachedText))
        {
             gv = cachedGlyphVector;
        }
        else if (g2d != null && text != null)
        {
            g2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, 
                    RenderingHints.VALUE_TEXT_ANTIALIAS_LCD_HRGB);
            FontRenderContext frc = g2d.getFontRenderContext();
            gv = gFont.layoutGlyphVector(frc, text.toCharArray(), 0, 
                text.length(), Font.LAYOUT_LEFT_TO_RIGHT);
            cachedGlyphVector = gv;
            cachedText = text;
            assert(dumpGV(gv, text));
        }
        else 
        {
            assert(debugPrint("Null g2d"));
        }
        return gv;
    }
    
    protected Rectangle getBounds(Shape a)
    {
      return (a instanceof Rectangle) ? (Rectangle)a : a.getBounds();
    }
    
    protected void assertPrint(String s) 
    {
        assert(debugPrint(s));
    }
    
    protected boolean debugPrint(String s)
    {
        debug = true;
        System.out.println(s);
        return true;
    }
    /**
     * debug dump of a glyph vector
     */
    protected boolean dumpGV(GlyphVector gv, String text)
    {
      DecimalFormat format = new DecimalFormat("####.#");
      int glyphCount = gv.getNumGlyphs();
      StringBuffer charIds = new StringBuffer();
      StringBuffer charCodes = new StringBuffer();
      StringBuffer glyphIds = new StringBuffer();
      StringBuffer glyphPos1 = new StringBuffer();
      StringBuffer glyphPos2 = new StringBuffer();
      for (int i = 0; i < glyphCount; i++)
      {
        if ((i % 10) == 0)
        {
          if (i > 0)
          {
            System.out.println(glyphIds.toString());
            System.out.println(glyphPos1.toString());
            System.out.println(glyphPos2.toString());
            System.out.println(charIds.toString());
            System.out.println(charCodes.toString());
            charIds.delete(0, charIds.length());
            charCodes.delete(0, charCodes.length());
            glyphIds.delete(0, glyphIds.length());
            glyphPos1.delete(0, glyphPos1.length());
            glyphPos2.delete(0, glyphPos2.length());
          }
        }
        else
        {
          charIds.append('\t');
          charCodes.append('\t');
          glyphIds.append('\t'); 
          glyphPos1.append('\t');
          glyphPos2.append('\t');
        }
        glyphIds.append(i);
        Point2D gBounds = gv.getGlyphPosition(i);
        glyphPos2.append(format.format(gBounds.getX()));
        charIds.append(gv.getGlyphCharIndex(i));
        charCodes.append(Integer.toHexString(Character.codePointAt(text, gv.getGlyphCharIndex(i))));
        charCodes.append("(");
        charCodes.append(text.charAt(gv.getGlyphCharIndex(i)));
        charCodes.append(")");
      }
      System.out.println(glyphIds.toString());
      System.out.println(glyphPos1.toString());
      System.out.println(glyphPos2.toString());
      System.out.println(charIds.toString());
      System.out.println(charCodes.toString());
      return true;
    }

    /* (non-Javadoc)
     * @see javax.swing.text.GlyphView.GlyphPainter#paint(javax.swing.text.GlyphView, java.awt.Graphics, java.awt.Shape, int, int)
     */
    @Override
    public void paint(GlyphView v, Graphics g, Shape a, int p0, int p1)
    {
        this.paint((View)v, g, a, p0, p1);
    }
    
    public View getView()
    {
        return view;
    }
    public GraphiteAwtFont getFont()
    {
        return gFont;
    }
}
