/*-----------------------------------------------------------------------------
Copyright (C) 2005,2007 Keith Stribley

Distributable under the terms of either the Common Public License or the
GNU Lesser General Public License, as specified in the LICENSING.txt file.

File: 
Responsibility: Keith Stribley
Last reviewed: Not yet.

Description:
Provides a simple JNI bridge between graphite and java
-----------------------------------------------------------------------------*/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <assert.h>

#include <jni.h>

#include <graphite/GrClient.h>
#include <graphite/ITextSource.h>
#include <graphite/IGrJustifier.h>
#include <graphite/IGrEngine.h>
#include <graphite/SegmentAux.h>
#include <graphite/Font.h>
#include <graphite/Segment.h>
#include <graphite/GraphiteProcess.h>
#include <graphite/FileFont.h>

#include "GrUtfTextSrc.h"

#include "org_sil_graphite_GraphiteFont.h"


typedef std::pair< gr::GlyphIterator, gr::GlyphIterator > GlyphRange;

struct SegmentHolder
{
public:
	SegmentHolder() : m_pSegment(NULL), m_pLayout(NULL), m_pTextSrc(NULL) {}
	~SegmentHolder()
	{
		delete m_pSegment;
		delete m_pLayout;
		delete m_pTextSrc;
		m_pSegment = NULL;
		m_pLayout = NULL;
		m_pTextSrc = NULL;
	}
	void setSegment(gr::Segment * pSegment) { m_pSegment = pSegment; }
	gr::Segment * getSegment(void) const { return m_pSegment; }
	void setLayout(gr::LayoutEnvironment * pLayout) { m_pLayout = pLayout; }
	gr::LayoutEnvironment * getLayout(void) const { return m_pLayout; }
	void setTextSrc(GrUtfTextSrc * pTextSrc) { m_pTextSrc = pTextSrc; }
	GrUtfTextSrc * getTextSrc(void) const { return m_pTextSrc; }
private:
	gr::Segment * m_pSegment;
	gr::LayoutEnvironment * m_pLayout;
	GrUtfTextSrc * m_pTextSrc;
};

#ifdef _DEBUG

#ifndef DEBUG_STR
#define DEBUG_STR(A) printf((A))
#define DEBUG_STR1(A,b) printf((A),(b))
#define DEBUG_STR2(A,b,c) printf((A),(b),(c))
#define DEBUG_STR3(A,b,c,d) printf((A),(b),(c),(d))
#define DEBUG_STR4(A,b,c,d,e) printf((A),(b),(c),(d),(e))
#define DEBUG_STR5(A,b,c,d,e,f) printf((A),(b),(c),(d),(e),(f))
#endif

#else

#ifndef DEBUG_STR
#define DEBUG_STR 
#define DEBUG_STR1 
#define DEBUG_STR2 
#define DEBUG_STR3 
#define DEBUG_STR4
#define DEBUG_STR5
#endif

#endif

JNIEXPORT jint JNICALL
JNI_OnLoad(JavaVM *vm, void *reserved)
{
	return JNI_VERSION_1_4;
}

/*
 * Class:     org_sil_graphite_GraphiteFont
 * Method:    createGraphiteFont
 * Signature: (Ljava/lang/String;FI)J
 */
JNIEXPORT jlong JNICALL Java_org_sil_graphite_GraphiteFont_createGraphiteFont
  (JNIEnv * env, jobject, jstring fontFileName, jfloat pointSize, jint dpi)
{
	const char * fileName = env->GetStringUTFChars(fontFileName, NULL);
	FILE	* file = fopen(fileName, "rb");
	gr::FileFont * fileFont = NULL;
	if (file)
	{
		fileFont = new gr::FileFont(file, pointSize, dpi, dpi);
		if (!fileFont->isValid())
		{
			DEBUG_STR("graphitejni:Invalid font!");
			delete fileFont;
			fileFont = NULL;
		}
	}
	env->ReleaseStringUTFChars(fontFileName, fileName);
	jlong fontHandle = reinterpret_cast<jlong>(fileFont);
	return fontHandle;
}

/*
 * Class:		 org_sil_graphite_GraphiteFont
 * Method:		isGraphiteFont
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_org_sil_graphite_GraphiteFont_isGraphiteFont
	(JNIEnv *, jobject, jlong fontHandle)
{
	gr::FileFont * fileFont = reinterpret_cast<gr::FileFont *>(fontHandle);
	jboolean isGraphite = false;
	if (fileFont) 
	{
		if(fileFont->isValid())
		  isGraphite = fileFont->fontHasGraphiteTables();
	}
	return isGraphite;
}

/*
 * Class:		 org_sil_graphite_GraphiteFont
 * Method:		createSegment
 * Signature: (J[C)J
 */
JNIEXPORT jlong JNICALL Java_org_sil_graphite_GraphiteFont_createSegment
	(JNIEnv *env, jobject, jlong fontHandle, jcharArray text, jboolean rtl)
{
	SegmentHolder * holder = new SegmentHolder;
	gr::FileFont * fileFont = reinterpret_cast<gr::FileFont *>(fontHandle);
	jlong segmentHandle = 0;
	holder->setTextSrc(new GrUtfTextSrc());
	if (!holder->getTextSrc())
	{
		delete holder;
		return segmentHandle;
	}
	holder->setLayout(new gr::LayoutEnvironment());
	if (!holder->getLayout())
	{
		delete holder;
		return segmentHandle;
	}
	int size = env->GetArrayLength(text);
	jboolean isCopy;
	gr::utf16 * pText = env->GetCharArrayElements(text, &isCopy);
	assert(pText);
    bool isRtl = (rtl) ? true : false;
	holder->getTextSrc()->setText(pText, size);
    holder->getTextSrc()->setRightToLeft(isRtl);
	// don't keep any changes to the array - there shouldn't be any
	env->ReleaseCharArrayElements(text, pText, JNI_ABORT);
	holder->getLayout()->setEndOfLine(false);
	holder->getLayout()->setStartOfLine(false);
	holder->getLayout()->setDumbFallback(true);
	holder->getLayout()->setJustifier(NULL);
	holder->setSegment(new gr::RangeSegment(fileFont, holder->getTextSrc(), 
											  holder->getLayout(), 0, size));
	if (holder->getSegment())
	{
		segmentHandle = reinterpret_cast<jlong>(holder);
		int numGlyphs = 0;
		GlyphRange gRange = holder->getSegment()->glyphs();
		for (gr::GlyphIterator gi = gRange.first; gi != gRange.second; ++gi)
			numGlyphs++;
		DEBUG_STR2("created segment %lx with %d glyphs\n", static_cast<long>(segmentHandle), numGlyphs);
	}
	else
	{
		delete holder;
		holder = NULL;
	}
	return segmentHandle;
}

/*
 * Class:		 org_sil_graphite_GraphiteFont
 * Method:		countGlyphs
 * Signature: (JJ)I
 */
JNIEXPORT jint JNICALL Java_org_sil_graphite_GraphiteFont_countGlyphs
	(JNIEnv *, jobject, jlong fontHandle, jlong segmentHandle)
{
	SegmentHolder * holder = reinterpret_cast<SegmentHolder *>(segmentHandle);
	assert(holder);
	assert(holder->getSegment());
	GlyphRange gRange = holder->getSegment()->glyphs();
	//jint numGlyphs = (gRange.second - gRange.first);
	jint numGlyphs = 0;
	for (gr::GlyphIterator gi = gRange.first; gi != gRange.second; ++gi)
		numGlyphs++;
    DEBUG_STR2("Segment %lx has %d glyphs\n", static_cast<long>(segmentHandle), numGlyphs);
	return numGlyphs;
}

/*
 * Class:		 org_sil_graphite_GraphiteFont
 * Method:		getGlyphArray
 * Signature: (JJ[I[F[I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_sil_graphite_GraphiteFont_getGlyphArray
	(JNIEnv * env, jobject, jlong fontHandle, jlong segmentHandle, 
	 jintArray gidArray, jfloatArray posArray, jintArray charIndexArray)
{
	jboolean success = false;
	SegmentHolder * holder = reinterpret_cast<SegmentHolder *>(segmentHandle);
	gr::FileFont * fileFont = reinterpret_cast<gr::FileFont *>(fontHandle);
	assert(fileFont);
	assert(holder);
	if (!holder->getSegment()) return success;
	float advanceWidth = holder->getSegment()->advanceWidth();
	GlyphRange gRange = holder->getSegment()->glyphs();
	jboolean isCopy;
	//jint numGlyphs = gRange.second - gRange.first;
	
	//assert(env->GetArrayLength(gidArray) == (numGlyphs));
	//assert(env->GetArrayLength(posArray) == ((numGlyphs) * 2 + 2));
	//assert(env->GetArrayLength(charIndexArray) == (numGlyphs));
	jint * gid = env->GetIntArrayElements(gidArray, &isCopy);
	jfloat * pos = env->GetFloatArrayElements(posArray, &isCopy);
	jint * charIndex = env->GetIntArrayElements(charIndexArray, &isCopy);
	
	gr::GlyphIterator gi = gRange.first;
	jint i = 0;
	jint prevChar = 0;
	DEBUG_STR2("getGlyphArray for segment %lx width %f\n", static_cast<long>(segmentHandle), advanceWidth);
	while (gi != gRange.second)
	{
		gr::GlyphInfo info = (*gi);
		gid[i] = info.glyphID();
		pos[2 * i] = info.origin();
		pos[(2 * i) + 1] = - info.yOffset();
		if (info.insertBefore())
		{
			charIndex[i] = info.firstChar();
			prevChar = charIndex[i];
		}
		else
		{
			charIndex[i] = prevChar;
		}
		DEBUG_STR5("%d:x=%f(%f, %f),c=%d\n", i, pos[2 * i], info.origin(), info.yOffset(), charIndex[i]);
		++i;
		++gi;
	}
	// assign last point to specify advance of the whole array
	// position arrays must be one bigger than what countGlyphs() returned
	pos[2 * i] = advanceWidth;
	pos[2 * i + 1] = 0.0;
	env->ReleaseIntArrayElements(gidArray, gid, 0);
	env->ReleaseFloatArrayElements(posArray, pos, 0);
	env->ReleaseIntArrayElements(charIndexArray, charIndex, 0);
	success = true;
	return success;
}

/*
 * Class:     org_sil_graphite_GraphiteFont
 * Method:    getGlyphInfo
 * Signature: (JJI[Z[I[F)Z
 */
JNIEXPORT jboolean JNICALL Java_org_sil_graphite_GraphiteFont_getGlyphInfo
  (JNIEnv * env, jobject, jlong fontHandle, jlong segmentHandle, jint gid, 
   jbooleanArray boolsArray, jintArray intsArray, jfloatArray floatsArray)
{
	jboolean success = false;
	SegmentHolder * holder = reinterpret_cast<SegmentHolder *>(segmentHandle);
	gr::FileFont * fileFont = reinterpret_cast<gr::FileFont *>(fontHandle);
	assert(fileFont);
	assert(holder);
	if (!holder->getSegment()) return success;
	GlyphRange gRange = holder->getSegment()->glyphs();
	// is the gid valid?
	if ((gRange.second - gRange.first) < gid)
		return success;
	gr::GlyphIterator gi = gRange.first + gid;
	gr::GlyphInfo & glyphInfo = *gi;
	
	// this is a bit messy, but it has the advantage that you don't need to 
	// use GetFieldID or GetMethodID, which you would need if we used a java class
	jboolean isCopy;
	jboolean * bools = env->GetBooleanArrayElements(boolsArray, &isCopy);
	jint * ints = env->GetIntArrayElements(intsArray, &isCopy);
	jfloat * floats = env->GetFloatArrayElements(floatsArray, &isCopy);
	
	bools[org_sil_graphite_GraphiteFont_GI_BOOL_IS_SPACE] = glyphInfo.isSpace();
	bools[org_sil_graphite_GraphiteFont_GI_BOOL_INSERT_BEFORE] = glyphInfo.insertBefore();
	bools[org_sil_graphite_GraphiteFont_GI_BOOL_IS_ATTACHED] = glyphInfo.isAttached();
	bools[org_sil_graphite_GraphiteFont_GI_BOOL_ERRONEOUS] = glyphInfo.erroneous();
	ints[org_sil_graphite_GraphiteFont_GI_INT_GLYPH_ID] = glyphInfo.glyphID();
	ints[org_sil_graphite_GraphiteFont_GI_INT_PSEUDO_GLYPH_ID] = glyphInfo.pseudoGlyphID();
	ints[org_sil_graphite_GraphiteFont_GI_INT_LOGICAL_INDEX] = static_cast<jint>(glyphInfo.logicalIndex());
	ints[org_sil_graphite_GraphiteFont_GI_INT_FIRST_CHAR] = glyphInfo.firstChar();
	ints[org_sil_graphite_GraphiteFont_GI_INT_LAST_CHAR] = glyphInfo.lastChar();
	ints[org_sil_graphite_GraphiteFont_GI_INT_DIRECTIONALITY] = glyphInfo.directionality();
	ints[org_sil_graphite_GraphiteFont_GI_INT_DIRECTION_LEVEL] = glyphInfo.directionLevel();
	ints[org_sil_graphite_GraphiteFont_GI_INT_BREAK_WEIGHT] = glyphInfo.breakweight();
	floats[org_sil_graphite_GraphiteFont_GI_FLOAT_ORIGIN] = glyphInfo.origin();
	floats[org_sil_graphite_GraphiteFont_GI_FLOAT_ADVANCE_WIDTH] = glyphInfo.advanceWidth();
	floats[org_sil_graphite_GraphiteFont_GI_FLOAT_ADVANCE_HEIGHT] = glyphInfo.advanceHeight();
	floats[org_sil_graphite_GraphiteFont_GI_FLOAT_Y_OFFSET] = glyphInfo.yOffset();
	floats[org_sil_graphite_GraphiteFont_GI_FLOAT_BB_LEFT] = glyphInfo.bb().left;
	floats[org_sil_graphite_GraphiteFont_GI_FLOAT_BB_RIGHT] = glyphInfo.bb().right;
	floats[org_sil_graphite_GraphiteFont_GI_FLOAT_BB_TOP] = glyphInfo.bb().top;
	floats[org_sil_graphite_GraphiteFont_GI_FLOAT_BB_BOTTOM] = glyphInfo.bb().bottom;
	
	env->ReleaseBooleanArrayElements(boolsArray, bools, 0);
	env->ReleaseIntArrayElements(intsArray, ints, 0);
	env->ReleaseFloatArrayElements(floatsArray, floats, 0);
	success = true;
	return success;
}
 
/*
 * Class:		 org_sil_graphite_GraphiteFont
 * Method:		destroySegment
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_org_sil_graphite_GraphiteFont_destroySegment
	(JNIEnv *, jobject, jlong segmentHandle)
{
	SegmentHolder * holder = reinterpret_cast<SegmentHolder *>(segmentHandle);
	assert(holder);
	DEBUG_STR1("destroySegment %lx\n", static_cast<long>(segmentHandle));
	delete holder;
}

/*
 * Class:		 org_sil_graphite_GraphiteFont
 * Method:		destroyFont
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_org_sil_graphite_GraphiteFont_destroyFont
	(JNIEnv *, jobject, jlong fontHandle)
{
	gr::FileFont * fileFont = reinterpret_cast<gr::FileFont *>(fontHandle);
	assert(fileFont);
	delete fileFont;
	fileFont = NULL;
}

/*
 * Class:     org_sil_graphite_GraphiteFont
 * Method:    getAscent
 * Signature: (J)F
 */
JNIEXPORT jfloat JNICALL Java_org_sil_graphite_GraphiteFont_getAscent
  (JNIEnv * env, jobject, jlong fontHandle)
{
	gr::FileFont * fileFont = reinterpret_cast<gr::FileFont *>(fontHandle);
	assert(fileFont);
	jfloat size = fileFont->ascent();
	return size;
}

/*
 * Class:     org_sil_graphite_GraphiteFont
 * Method:    getDescent
 * Signature: (J)F
 */
JNIEXPORT jfloat JNICALL Java_org_sil_graphite_GraphiteFont_getDescent
  (JNIEnv * env, jobject, jlong fontHandle)
{
	gr::FileFont * fileFont = reinterpret_cast<gr::FileFont *>(fontHandle);
	assert(fileFont);
	jfloat size = fileFont->descent();
	return size;
}

/*
 * Class:     org_sil_graphite_GraphiteFont
 * Method:    getHeight
 * Signature: (J)F
 */
JNIEXPORT jfloat JNICALL Java_org_sil_graphite_GraphiteFont_getHeight
  (JNIEnv * env, jobject, jlong fontHandle)
{
	gr::FileFont * fileFont = reinterpret_cast<gr::FileFont *>(fontHandle);
	assert(fileFont);
	jfloat size = fileFont->height();
	return size;
}
